#include "cmdu_validate.h"

#include <1905_tlvs.h>
#include <cmdu.h>
#include <easy/bufutil.h>
#include <easy/utils.h>
#include <easymesh.h>
#include <stdlib.h>

#include "utils/debug.h"


#define ATTR_MSG_TYPE	(0x1022)
#define MSG_TYPE_M2		(0x05)
static int validate_wsc_m2(uint8_t *m2, uint16_t m2_size)
{
	uint8_t *data;
	uint8_t *m2_end;
	bool ret = -1;

	if (!m2 || !m2_size)
		return -1;

	data = m2;
	m2_end = m2 + m2_size;

	while (labs(data - m2) < m2_size - 4) {
		uint16_t attr_type;
		uint16_t attr_len;

		attr_type = buf_get_be16(data);
		data += 2;
		attr_len = buf_get_be16(data);
		data += 2;

		if (data + attr_len > m2_end) {
			dbg("%s: parse_wsc_m2 failed\n", __func__);
			ret = -1;
			break;
		}

		if (attr_type == ATTR_MSG_TYPE) {
			if (attr_len != 1) {
				ret = -1;
				break;
			}
			if (*data == MSG_TYPE_M2)
				ret = 0;
		}

		data += attr_len;
	}

	/* true if msg type is M2 & data never goes OOB */
	return ret;
}

/* Check WSC TLV (containing M2) */
static int check_wsc_tlv(struct tlv *t)
{
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);
	if (!tlv_len)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	return validate_wsc_m2(tv_data, tlv_len);
}

static int check_serialized_tlv(struct tlv *t, uint16_t len)
{
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);

	if (tlv_len != len)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	return 0;
}

/* Check AP Radio Indentifier TLV */
static int check_ap_radio_identifier_tlv(struct tlv *t)
{
	/* bssid (6 bytes) */
	return check_serialized_tlv(t, 6);
}

/* Check Default 802.1Q Settings TLV */
static int check_default_11q_settings_tlv(struct tlv *t)
{
	/* pvid + pcp (3 bytes) */
	return check_serialized_tlv(t,
				sizeof(struct tlv_default_8021q_settings));
}

/* Check Traffic Separation Policy TLV */
static int check_traffic_separation_policy_tlv(struct tlv *t)
{
	int i, offset = 0;
	uint8_t num_ssid;
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);
	/* at least 1 byte: num_ssid */
	if (tlv_len < 1)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	num_ssid = tv_data[offset++];

	for (i = 0; i < num_ssid; i++) {
		int ssid_len;

		/* ssid_len (1 byte) */
		if (offset + 1 > tlv_len)
			return -1;
		ssid_len = tv_data[offset++];

		/* ssid (ssid_len bytes) */
		if (offset + ssid_len > tlv_len)
			return -1;
		offset += ssid_len;

		/* vid (2 bytes) */
		if (offset + 2 > tlv_len)
			return -1;
		offset += 2;
	}

	return 0;
}

/* TODO: Remove once fix in ieee1905 done (now using common M1/M2 CMDU type) */
bool validate_ap_autoconfig_wsc(struct cmdu_buff *cmdu, struct tlv *tv[][TLV_MAXNUM], uint8_t profile)
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_AP_RADIO_IDENTIFIER,
				.present = TLV_PRESENT_ONE,
				.minlen = 6, /* bssid */
				.maxlen = 6,
		},
		[1] = { .type = TLV_TYPE_WSC,
				.present = TLV_PRESENT_MORE, /* M2 - more; M1 - one */
		},
		[2] = { .type = MAP_TLV_DEFAULT_8021Q_SETTINGS,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 3, /* tlv_default_8021q_settings */
		},
		[3] = { .type = MAP_TLV_TRAFFIC_SEPARATION_POLICY,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* tlv_traffic_sep_policy: num_ssid */
		},
#if (EASYMESH_VERSION >= 6)
		[4] = { .type = MAP_TLV_AP_MLD_CONFIG,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 1,
		},
		[5] = { .type = MAP_TLV_BACKHAUL_STA_MLD_CONFIG,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 1,
		},
#endif

	};
	int num = 0;
	int ret;
	int policy_len = 4;

	trace("%s |" MACFMT "|CMDU: ap autoconfig WSC\n",
	      __func__, MAC2STR(cmdu->origin));

#if (EASYMESH_VERSION >= 6)
	policy_len = 6;
#endif
	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, policy_len);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[1][0]) {
		dbg("%s: Missing one or more mandatory TLV!\n", __func__);
		return false;
	}

	/* Parse AP Radio Identifier TLV */
	if (check_ap_radio_identifier_tlv(tv[0][0]))
		return false;

	/* Parse WSC TLVs (containing M2) */
	while (num < TLV_MAXNUM && tv[1][num]) {
		if (check_wsc_tlv(tv[1][num]))
			return false;
		num++;
	}

	/* Parse Default 802.1Q Settings TLV */
	if (tv[2][0] && check_default_11q_settings_tlv(tv[2][0]))
		return false;

	/* Parse Traffic Separation Policy TLV */
	if (tv[3][0] && check_traffic_separation_policy_tlv(tv[3][0]))
		return false;

	return true;
}

#ifdef EASYMESH_VENDOR_EXT
bool validate_vendor_specific(struct cmdu_buff *cmdu, struct tlv *tlvs[][TLV_MAXNUM], size_t tlvs_size, uint8_t profile)
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_VENDOR_SPECIFIC,
				.present = TLV_PRESENT_OPTIONAL_MORE,
				.minlen = 3,
		},
	};
	int ret;

	trace("%s |" MACFMT "|CMDU: vendor specific\n",
		  __func__, MAC2STR(cmdu->origin));

	ret = cmdu_parse_tlvs(cmdu, tlvs, a_policy, 1);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tlvs[0][0]) {
		dbg("%s: Missing one or more mandatory TLV!\n", __func__);
		return false;
	}

	return true;
}
#endif
